@@ -8,14 +8,32 @@ use Regexp::RegGrp;
 
 # -----------------------------------------------------------------------------
 
-our $VERSION = '1.000';
+our $VERSION = '1.001_001';
 
-our @TAGS = (
+our @BOOLEAN_ACCESSORS = (
+    'remove_comments',
+    'remove_newlines',
+    'no_compress_comment',
+    'html5',
+);
+
+our @JAVASCRIPT_OPTS    = ( 'minify', 'shrink', 'base62' );
+our @CSS_OPTS           = ( 'minify', 'pretty' );
+
+our $REQUIRED_JAVASCRIPT_PACKER = '1.002001';
+our $REQUIRED_CSS_PACKER        = '1.000001';
+
+our @SAVE_SPACE_ELEMENTS = (
     'a', 'abbr', 'acronym', 'address', 'b', 'bdo', 'big', 'button', 'cite',
     'del', 'dfn', 'em', 'font', 'i', 'input', 'ins', 'kbd', 'label', 'q',
     's', 'samp', 'select', 'small', 'strike', 'strong', 'sub', 'sup', 'u', 'var'
 );
 
+our @VOID_ELEMENTS = (
+        'area', 'base', 'br', 'col', 'command', 'embed', 'hr', 'img', 'input',
+        'keygen', 'link', 'meta', 'param', 'source', 'track', 'wbr'
+);
+
 # Some regular expressions are from HTML::Clean
 
 our $COMMENT        = '((?>\s*))(<!--(?:(?:[^#\[]|(?! google_ad_section_)).*?)?-->)((?>\s*))';
@@ -74,7 +92,7 @@ our $WHITESPACES    = [
 
 our $NEWLINES_TAGS = [
     {
-        regexp      => '(\s*)(<\s*\/?\s*(?:' . join( '|', @TAGS ) . ')[^>]*>)(\s*)',
+        regexp      => '(\s*)(<\s*\/?\s*(?:' . join( '|', @SAVE_SPACE_ELEMENTS ) . ')\b[^>]*>)(\s*)',
         replacement => sub {
             return sprintf( '%s%s%s', $_[0]->{submatches}->[0] ? ' ' : '', $_[0]->{submatches}->[1], $_[0]->{submatches}->[2] ? ' ' : '' );
         },
@@ -105,20 +123,73 @@ our $NEWLINES = [
     }
 ];
 
+our @REGGRPS        = ( 'newlines', 'newlines_tags', 'whitespaces', 'void_elements' );
+
+our $GLOBAL_REGGRP  = 'global';
+
+##########################################################################################
+
+{
+    no strict 'refs';
+
+    foreach my $field ( @BOOLEAN_ACCESSORS ) {
+        next if defined *{ __PACKAGE__ . '::' . $field }{CODE};
+
+        *{ __PACKAGE__ . '::' . $field} = sub {
+            my ( $self, $value ) = @_;
+
+            $self->{'_' . $field} = $value ? 1 : undef if ( defined( $value ) );
+
+            return $self->{'_' . $field};
+        };
+    }
+
+    foreach my $reggrp ( @REGGRPS, $GLOBAL_REGGRP ) {
+        next if defined *{ __PACKAGE__ . '::reggrp_' . $reggrp }{CODE};
+
+        *{ __PACKAGE__ . '::reggrp_' . $reggrp } = sub {
+            my ( $self ) = shift;
+
+            return $self->{ '_reggrp_' . $reggrp };
+        };
+    }
+}
+
+sub do_javascript {
+    my ( $self, $value ) = @_;
+
+    if ( defined( $value ) ) {
+        if ( grep( $value eq $_, @JAVASCRIPT_OPTS ) ) {
+            $self->{_do_javascript} = $value;
+        }
+        elsif ( ! $value ) {
+            $self->{_do_javascript} = undef;
+        }
+    }
+
+    return $self->{_do_javascript};
+}
+
+sub do_stylesheet {
+    my ( $self, $value ) = @_;
+
+    if ( defined( $value ) ) {
+        if ( grep( $value eq $_, @CSS_OPTS ) ) {
+            $self->{_do_stylesheet} = $value;
+        }
+        elsif ( ! $value ) {
+            $self->{_do_stylesheet} = undef;
+        }
+    }
+
+    return $self->{_do_stylesheet};
+}
+
 sub init {
     my $class = shift;
     my $self  = {};
 
-    eval {
-        require JavaScript::Packer;
-    };
-    $self->{can_do_javascript}  = $@ ? 0 : 1;
-    $self->{javascript_packer}  = undef;
-    eval {
-        require CSS::Packer;
-    };
-    $self->{can_do_stylesheet}  = $@ ? 0 : 1;
-    $self->{css_packer}         = undef;
+    bless( $self, $class );
 
     $self->{whitespaces}->{reggrp_data}   = $WHITESPACES;
     $self->{newlines}->{reggrp_data}      = $NEWLINES;
@@ -140,25 +211,17 @@ sub init {
         {
             regexp      => $COMMENT,
             replacement => sub {
-                my $opts            = $_[0]->{opts} || {};
-                my $remove_comments = _get_opt( $opts, 'remove_comments' );
-                my $remove_newlines = _get_opt( $opts, 'remove_newlines' );
-
-                return $remove_comments ? (
-                    $remove_newlines ? ' ' : (
+                return $self->remove_comments() ? (
+                    $self->remove_newlines() ? ' ' : (
                         ( $_[0]->{submatches}->[0] =~ /\n/s or $_[0]->{submatches}->[2] =~ /\n/s ) ? "\n" : ''
                     )
                 ) : '<!--~' . $_[0]->{store_index} . '~-->';
             },
             store => sub {
-                my $opts            = $_[0]->{opts} || {};
-                my $remove_comments = _get_opt( $opts, 'remove_comments' );
-                my $remove_newlines = _get_opt( $opts, 'remove_newlines' );
-
-                my $ret = $remove_comments ? '' : (
-                     ( ( not $remove_newlines and $_[0]->{submatches}->[0] =~ /\n/s ) ? "\n" : '' ) .
+                my $ret = $self->remove_comments() ? '' : (
+                     ( ( not $self->remove_newlines() and $_[0]->{submatches}->[0] =~ /\n/s ) ? "\n" : '' ) .
                      $_[0]->{submatches}->[1] .
-                     ( ( not $remove_newlines and $_[0]->{submatches}->[2] =~ /\n/s ) ? "\n" : '' )
+                     ( ( not $self->remove_newlines() and $_[0]->{submatches}->[2] =~ /\n/s ) ? "\n" : '' )
                 );
 
                 return $ret;
@@ -171,24 +234,25 @@ sub init {
             },
             store => sub {
                 my ( $opening, undef, $content, $closing )  = @{$_[0]->{submatches}};
-                my $opts                                    = $_[0]->{opts} || {};
 
                 if ( $content ) {
-                    if ( $opening =~ /<\s*script[^>]*(?:java|ecma)script[^>]*>/ and $self->{javascript_packer} ) {
-                        my $do_javascript = _get_opt( $opts, 'do_javascript' );
-                        if ( $do_javascript ) {
-                            my $no_cdata = _get_opt( $opts, 'no_cdata' );
-                            $self->{javascript_packer}->minify( \$content, { compress => $do_javascript } );
-                            unless ( $no_cdata ) {
+                    my $opening_re = '<\s*script' . ( $self->html5() ? '[^>]*>' : '[^>]*(?:java|ecma)script[^>]*>' );
+
+                    if ( $opening =~ /$opening_re/i ) {
+                        $opening =~ s/ type="(text\/)?(java|ecma)script"//i if ( $self->html5() );
+
+                        if ( $self->javascript_packer() and $self->do_javascript() ) {
+                            $self->javascript_packer()->minify( \$content, { compress => $self->do_javascript() } );
+
+                            unless ( $self->html5() ) {
                                 $content = '/*<![CDATA[*/' . $content . '/*]]>*/';
                             }
                         }
                     }
-                    elsif ( $opening =~ /<\s*style[^>]*text\/css[^>]*>/ and $self->{css_packer} ) {
-                        my $do_stylesheet = _get_opt( $opts, 'do_stylesheet' );
-                        if ( $do_stylesheet ) {
-                            $self->{css_packer}->minify( \$content, { compress => $do_stylesheet } );
-                            $content = "\n" . $content if ( $do_stylesheet eq 'pretty' );
+                    elsif ( $opening =~ /<\s*style[^>]*text\/css[^>]*>/ ) {
+                        if ( $self->css_packer() and $self->do_stylesheet() ) {
+                            $self->css_packer()->minify( \$content, { compress => $self->do_stylesheet() } );
+                            $content = "\n" . $content if ( $self->do_stylesheet() eq 'pretty' );
                         }
                     }
                 }
@@ -196,12 +260,8 @@ sub init {
                     $content = '';
                 }
 
-                # I don't like this, but
-                # $self->{whitespaces}->{reggrp}->exec( \$opening );
-                # will not work. It isn't initialized jet.
-                # If someone has a better idea, please let me know
-                $self->_process_wrapper( 'whitespaces', \$opening );
-                $self->_process_wrapper( 'whitespaces', \$closing );
+                $self->reggrp_whitespaces()->exec( \$opening );
+                $self->reggrp_whitespaces()->exec( \$closing );
 
                 return $opening . $content . $closing;
             },
@@ -209,19 +269,27 @@ sub init {
         }
     ];
 
-    map {
-        $self->{$_}->{reggrp} = Regexp::RegGrp->new( { reggrp => $self->{$_}->{reggrp_data} } );
-    } ( 'newlines', 'newlines_tags', 'whitespaces' );
-
-    $self->{global}->{reggrp} = Regexp::RegGrp->new(
+    $self->{void_elements}->{reggrp_data} = [
         {
-            reggrp          => $self->{global}->{reggrp_data},
+            regexp      => '<\s*((?:' . join( '|', @VOID_ELEMENTS ) . ')\b[^>]*)\s*\/>',
+            replacement => sub {
+                return '<' . $_[0]->{submatches}->[0] . '>';
+            },
+            modifier    => 'ism'
+        }
+    ];
+
+    map {
+        $self->{ '_reggrp_' . $_ } = Regexp::RegGrp->new( { reggrp => $self->{$_}->{reggrp_data} } );
+    } @REGGRPS;
+
+    $self->{ '_reggrp_' . $GLOBAL_REGGRP } = Regexp::RegGrp->new(
+        {
+            reggrp          => $self->{$GLOBAL_REGGRP}->{reggrp_data},
             restore_pattern => qr/<!--~(\d+)~-->/
         }
     );
 
-    bless( $self, $class );
-
     return $self;
 }
 
@@ -247,7 +315,7 @@ sub minify {
         return undef;
     }
 
-    my $html    = \'';
+    my $html;
     my $cont    = 'void';
 
     if ( defined( wantarray ) ) {
@@ -260,84 +328,71 @@ sub minify {
         $html = ref( $input ) ? $input : \$input;
     }
 
-    if ( $self->{can_do_javascript} and not $self->{javascript_packer_isset} ) {
-        $self->{javascript_packer} = eval {
-            JavaScript::Packer->init();
-        };
-        $self->{javascript_packer_isset} = 1;
+    if ( ref( $opts ) eq 'HASH' ) {
+        foreach my $field ( @BOOLEAN_ACCESSORS ) {
+            $self->$field( $opts->{$field} ) if ( defined( $opts->{$field} ) );
+        }
+
+        $self->do_javascript( $opts->{do_javascript} ) if ( defined( $opts->{do_javascript} ) );
+        $self->do_stylesheet( $opts->{do_stylesheet} ) if ( defined( $opts->{do_stylesheet} ) );
     }
 
-    if ( $self->{can_do_stylesheet} and not $self->{css_packer_isset} ) {
-        $self->{css_packer} = eval {
-            CSS::Packer->init();
-        };
-        $self->{css_packer_isset} = 1;
-    }
-
-    if ( ref( $opts ) ne 'HASH' ) {
-        carp( 'Second argument must be a hashref of options! Using defaults!' ) if ( $opts );
-        $opts = {
-            remove_comments     => 0,
-            remove_newlines     => 0,
-            do_javascript       => '',  # minify, shrink, base62
-            do_stylesheet       => '',  # pretty, minify
-            no_compress_comment => 0,
-            no_cdata            => 0
-        };
-    }
-    else {
-        $opts->{remove_comments} = $opts->{remove_comments} ? 1 : 0;
-        $opts->{remove_newlines} = $opts->{remove_newlines} ? 1 : 0;
-        $opts->{do_javascript}   = (
-            grep( $opts->{do_javascript}, ( 'minify', 'shrink', 'base62' ) ) &&
-            $self->{javascript_packer}
-        ) ? $opts->{do_javascript} : '';
-
-        $opts->{do_stylesheet}  = (
-            grep( $opts->{do_stylesheet}, ( 'minify', 'pretty' ) ) &&
-            $self->{css_packer}
-        ) ? $opts->{do_stylesheet} : '';
-
-        $opts->{no_compress_comment}    = $opts->{no_compress_comment} ? 1 : 0;
-        $opts->{no_cdata}               = $opts->{no_cdata} ? 1 : 0;
-    }
-
-    if ( not $opts->{no_compress_comment} and ${$html} =~ /$PACKER_COMMENT/s ) {
+    if ( not $self->no_compress_comment() and ${$html} =~ /$PACKER_COMMENT/s ) {
         my $compress = $1;
         if ( $compress eq '_no_compress_' ) {
             return ( $cont eq 'scalar' ) ? ${$html} : undef;
         }
     }
 
-    $self->{global}->{reggrp}->exec( $html, $opts );
-    $self->{whitespaces}->{reggrp}->exec( $html, $opts );
-    if ( $opts->{remove_newlines} ) {
-        $self->{newlines_tags}->{reggrp}->exec( $html );
-        $self->{newlines}->{reggrp}->exec( $html );
+    $self->reggrp_global()->exec( $html );
+    $self->reggrp_whitespaces()->exec( $html );
+    if ( $self->remove_newlines() ) {
+        $self->reggrp_newlines_tags()->exec( $html );
+        $self->reggrp_newlines()->exec( $html );
+    }
+    if ( $self->html5() ) {
+        $self->reggrp_void_elements()->exec( $html );
     }
 
-    $self->{global}->{reggrp}->restore_stored( $html );
+    $self->reggrp_global()->restore_stored( $html );
 
     return ${$html} if ( $cont eq 'scalar' );
 }
 
-sub _get_opt {
-    my ( $opts_hash, $opt ) = @_;
+sub javascript_packer {
+    my $self = shift;
 
-    $opts_hash  ||= {};
-    $opt        ||= '';
+    unless ( $self->{_checked_javascript_packer} ) {
+        eval "use JavaScript::Packer $REQUIRED_JAVASCRIPT_PACKER;";
 
-    my $ret = '';
+        unless ( $@ ) {
+            $self->{_javascript_packer} = eval {
+                JavaScript::Packer->init();
+            };
+        }
 
-    $ret = $opts_hash->{$opt} if ( defined( $opts_hash->{$opt} ) );
+        $self->{_checked_javascript_packer} = 1;
+    }
 
-    return $ret;
+    return $self->{_javascript_packer};
 }
 
-sub _process_wrapper {
-    my ( $self, $reg_name, $in, $opts ) = @_;
+sub css_packer {
+    my $self = shift;
 
-    $self->{$reg_name}->{reggrp}->exec( $in, $opts );
+    unless ( $self->{_checked_css_packer} ) {
+        eval "use CSS::Packer $REQUIRED_CSS_PACKER;";
+
+        unless ( $@ ) {
+            $self->{_css_packer} = eval {
+                CSS::Packer->init();
+            };
+        }
+
+        $self->{_checked_css_packer} = 1;
+    }
+
+    return $self->{_css_packer};
 }
 
 1;
@@ -350,7 +405,7 @@ HTML::Packer - Another HTML code cleaner
 
 =head1 VERSION
 
-Version 1.000
+Version 1.001_001
 
 =head1 DESCRIPTION
 
@@ -403,7 +458,11 @@ If not set to a true value it is allowed to set a HTML comment that prevents the
 
     <!-- HTML::Packer _no_compress_ -->
 
-Is set by default.
+Is not set by default.
+
+=item html5
+
+If set to a true value closing slashes will be removed from void elements.
 
 =back
 
@@ -413,8 +472,8 @@ Merten Falk, C<< <nevesenin at cpan.org> >>
 
 =head1 BUGS
 
-Please report any bugs or feature requests to C<bug-html-packer at rt.cpan.org>, or through
-the web interface at L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=HTML-Packer>.  I will be notified, and then you'll
+Please report any bugs or feature requests through
+the web interface at L<https://github.com/nevesenin/html-packer-perl/issues>. I will be notified, and then you'll
 automatically be notified of progress on your bug as I make changes.
 
 =head1 SUPPORT
